(*

BWT ENGINE FOR BINARY DATA

Slow, but very simple.

(C) Implemented by Alexander Myasnikov

Web: www.darksoftware.narod.ru
E-mail: darksoftware@ya.ru

Warning: "@" char is EOL marker

*)


unit bwt;

interface


type
  TBwtChar   = integer;
  TBwtString = array of TBwtChar;
  TBwtBlock  = array of TBwtString;

procedure Sort(var block: array of byte; len: integer);
procedure ReverseSort(var block: array of byte; len: integer);

implementation

function ArrayShift(block: TBwtString): TbwtString;
var
  i: integer;
begin

  SetLength(Result, Length(Block));
  Result[0] := Block[High(Block)];

  for I := 1 to High(Result) do
  begin
    Result[i] := Block[i - 1];
  end;

end;

function RowCompare(check, current: TbwtString; limit: integer = 0): boolean;
var
  i: integer;
begin

  Result := False;

  if Length(current) = 0 then
  begin
    Result := True;
    exit;
  end;

  if Length(check) = 0 then
  begin
    Result := False;
    exit;
  end;

  for I := limit to High(check) do
  begin

    if Check[i] > Current[i] then
    begin
      Result := True;
      Exit;
    end;

    if Check[i] < Current[i] then
    begin
      Result := False;
      Exit;
    end;

  end;

end;

function GetMaxRow(var block: TBwtBlock; block_size: integer;
  limit: integer = 0): TbwtString;
var
  i, d: integer;
begin

  SetLength(Result, 0);

  for i := 0 to block_size - 1 do
  begin

    if RowCompare(block[i], Result, limit) then
    begin
      d := i;
      Result := block[d];
    end;

  end;

  SetLength(block[d], 0);

end;


procedure ArraySort(var block: TBwtBlock; block_size: integer; limit: integer = 0);
var
  TBlock: TBwtBlock;
  i: integer;
begin

  SetLength(tBlock, Length(Block));

  for I := 0 to block_size - 1 do
  begin
    tBlock[i] := Block[i];
  end;

  for I := block_size - 1 downto 0 do
  begin
    Block[i] := GetMaxRow(tBlock, block_size, limit);
  end;

  for I := 0 to block_size - 1 do
  begin
    SetLength(tBlock[i], 0);
  end;

  SetLength(tBlock, 0);

end;


procedure Sort(var block: array of byte; len: integer);
var
  i:  integer;
  BS: TBwtString;
  BB: TBWTBlock;
  block_size: integer;
begin

  block_size := len + 1;

  SetLength(BS, block_size);

  for I := 0 to block_size - 2 do
  begin
    BS[i] := block[i];
  end;

  BS[block_size - 1] := 256;

  SetLength(BB, block_size);

  BB[0] := BS;

  for I := 1 to block_size - 1 do
  begin
    BB[I] := ArrayShift(BB[i - 1]);
  end;


  ArraySort(BB, block_size);

  for I := 0 to block_size - 1 do
  begin
    if BB[i][block_size - 1] <> 256 then
      block[i] := BB[i][block_size - 1]
    else
      block[i] := byte('@');
  end;

end;




procedure ReverseSort(var block: array of byte; len: integer);
var
  i, n: integer;
  BS: TBwtString;
  BB: TBWTBlock;
  block_size: integer;
begin

  block_size := len;
  SetLength(BB, block_size);

  for I := 0 to len - 1 do
  begin
    SetLength(BB[i], len);

    for n := 0 to len - 1 do
      BB[i][n] := 0;
  end;

  SetLength(BS, len);
  for I := 0 to len - 1 do
  begin
    if block[i] = byte('@') then
      BS[i] := 256
    else
      BS[i] := Block[i];
  end;

  for n := len - 1 downto 0 do

  begin
    for i := 0 to len - 1 do
      BB[i][n] := BS[i];
    ArraySort(BB, block_size, n);

  end;

  for I := 0 to len - 1 do
  begin

    if BB[i][len - 1] = 256 then
    begin

      for n := 0 to len - 1 do
      begin
        if BB[i][n] = 256 then
          block[n] := byte('@')
        else
          block[n] := BB[i][n];
      end;

      break;
    end;

  end;

end;

end.
